import java.io.*;
import javax.xml.parsers.*;
import org.xml.sax.helpers.*;

import java.util.logging.*;
import java.util.*;

import org.xml.sax.*;

/** 
 * Parser for the alignment xml files. Out of one or more alignment files it builds a Context object.
 * 
 * @author darko
 *
 */
public class AlignmentParser {
	Logger logger = Logger.getLogger(AlignmentParser.class.getName());

	protected Context context;
	protected Alignment currentAlignment;

	public AlignmentParser() {
		this.context = new Context();
	}

	public AlignmentParser(String contextName) {
		this.context = new Context(contextName);
	}

	public AlignmentParser(Context context) {
		this.context = context;
	}

	public Context getContext() {
		return this.context;
	}

	public void setContext(Context context) {
		this.context = context;
	}

	public Context parse(String filename) {
		parseFile(new File(filename));
		return this.context;
	}

	public Context parse(File file) {
		parseFile(file);
		return this.context;
	}

	public Context parse(String... filenames) {
		for (String filename : filenames) {
			parseFile(new File(filename));
		}
		return this.context;
	}

	public Context parse(File... files) {
		for (File file : files) {
			parseFile(file);
		}
		return this.context;
	}

	private void parseFile(File file) {
		
		SAXParserFactory factory = SAXParserFactory.newInstance();
		factory.setValidating(true);
		try {
			SAXParser parser = factory.newSAXParser();
			parser.parse(file, new AlignmentHandler());

		} catch (Exception ex) {
			logger.severe(ex.toString());
			ex.printStackTrace();
			throw new RuntimeException();
		}

	}

	class AlignmentHandler
			extends DefaultHandler {

		@Override
		public void startElement(String uri, String localName, String name, Attributes attributes) throws SAXException {
			if (name.equals("linkList")) {
				parseLinkList(attributes);
			} else if (name.equals("linkGrp")) {
				parseLinkGrp(attributes);
			} else if (name.equals("link")) {
				parseLink(attributes);
			}
		}
	}

	private void parseLinkList(Attributes attributes) {
		String linkListID = attributes.getValue("id");
		if (linkListID == null) {
			logger.severe("Element 'linkList' is missing attribute 'id'");
			throw new RuntimeException();
		}
		if (context.alignments.containsKey(linkListID)) {
			logger.warning("Context already contains alignment '" + linkListID + "'");
			int i = 1;
			while (context.alignments.containsKey(linkListID + "_" + i)) {
				i++;
			}
			linkListID += ("_" + i);
			logger.warning("Renaming to '" + linkListID + "'");
		}
		currentAlignment = new Alignment(linkListID);
		context.alignments.put(linkListID, currentAlignment);
	}

	private void parseLinkGrp(Attributes attributes) {
		String domains = attributes.getValue("domains");
		if (domains == null) {
			logger.severe("Element 'linkGrp' is missing attribute 'domains'");
			throw new RuntimeException();
		}
		String[] corporaNames = domains.trim().split("\\s+");
		for (String corpusName : corporaNames) {
			Corpus corpus = context.corpora.get(corpusName);
			if (corpus == null) {
				corpus = new Corpus(corpusName);
				context.corpora.put(corpusName, corpus);
			}
			currentAlignment.corpora.add(corpus);
			corpus.alignments.add(currentAlignment);
		}
		currentAlignment.setN(corporaNames.length);
	}

	private void parseLink(Attributes attributes) {
		String xtargets = attributes.getValue("xtargets");
		if (xtargets == null) {
			logger.severe("Element 'link' is missing attribute 'xtargets'");
			throw new RuntimeException();
		}
		String[] linkGroups = xtargets.split(";", -1);
		if (linkGroups.length != currentAlignment.n) {
			logger.severe(linkGroups.length + "-way link was encountered, " + currentAlignment.n
					+ "-way link was expected");
			throw new RuntimeException();
		}
		Link link = new Link(xtargets);
		for (int i = 0; i < currentAlignment.n; i++) {
			String linkGroupString = linkGroups[i];
			String[] sentenceIDs = linkGroupString.trim().split("\\s+");
			Set<Sentence> linkGroup = new LinkedHashSet<Sentence>(sentenceIDs.length);
			Corpus currentCorpus = currentAlignment.corpora.get(i);
			if (sentenceIDs[0].isEmpty()) { // group has 0 sentences
				Sentence emptySentence = new Sentence();
				emptySentence.links.put(currentAlignment.name, link);
				linkGroup.add(emptySentence);
				currentCorpus.sentences.put(emptySentence.id, emptySentence);
			} else { // group has non-0 sentences
				for (String sentenceID : sentenceIDs) {
					Sentence sentence = currentCorpus.sentences.get(sentenceID);
					if (sentence == null) {
						sentence = new Sentence(sentenceID);
						currentCorpus.sentences.put(sentenceID, sentence);
					}
					Link previous = sentence.links.put(currentAlignment.name, link);
					linkGroup.add(sentence);
					if (previous != null) {
						logger.warning("Sentence '" + sentence.id + "' appears more than once in alignment '"
								+ currentAlignment.name + "'. Only last occurence will be kept.");
					}
				}
			}
			link.linkGroups.put(currentCorpus.name, linkGroup);
		}
		currentAlignment.links.add(link);
	}

}
